@@ -1,9 +1,15 @@
 #! /bin/sh
-# This script creates a fresh repository
+
+# This script is part of CVS Utilities, version @VERSION@.
+# This script discards local changes and non-CVS files, creating a clean
+# working directory.
+
+set -e
+
 echo "cvsco: cleaning up ..."
 cvsu --ignore --find --types '^.FCRUX' --batch "rm -rf" "$@"
-test $? -eq 0 || exit 1
+
 echo "cvsco: refetching missing files ..."
 cvsu --ignore --find --types 'U' --batch "cvs update" "$@"
-test $? -eq 0 || exit 1
+
 echo "cvsco: done"
