@@ -1,31 +1,78 @@
 #! /usr/bin/perl -w
 
 # cvsu - do a quick check to see what files are out of date.
+#
+# Copyright (C) 2000-2005  Pavel Roskin <proski@gnu.org>
 # Initially written by Tom Tromey <tromey@cygnus.com>
-# Rewritten by Pavel Roskin <proski@gnu.org>
+# Completely rewritten by Pavel Roskin <proski@gnu.org>
+#
+# This program is free software; you can redistribute it and/or modify
+# it under the terms of the GNU General Public License as published by
+# the Free Software Foundation; either version 2, or (at your option)
+# any later version.
+#
+# This program is distributed in the hope that it will be useful,
+# but WITHOUT ANY WARRANTY; without even the implied warranty of
+# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+# GNU General Public License for more details.
+#
+# You should have received a copy of the GNU General Public License
+# along with this program; if not, write to the Free Software
+# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
+# 02111-1307, USA.
 
-# $Id: cvsu.pl,v 1.11 2002/08/03 22:36:42 proski Exp $
 
 require 5.004;
 use Getopt::Long;
+use File::Basename;
 use Time::Local;
 use strict;
 
 use vars qw($list_types %messages %options @batch_list $batch_cmd
 	    $no_recurse $explain_type $find_mode $short_print
-	    $no_cvsignore $file $curr_dir
-	    @standard_ignores $ignore_rx);
+	    $no_cvsignore $nolinks $file $single_filename $curr_dir
+	    @common_ignores $ignore_rx %entries %subdirs %removed);
+
+use constant SUBDIR_FOUND => 1;
+use constant SUBDIR_CVS   => 2;
+
+# This list comes from the CVS manual.
+use constant STANDARD_IGNORES =>
+	('RCS', 'SCCS', 'CVS', 'CVS.adm', 'RCSLOG', 'cvslog.*', 'tags',
+	 'TAGS', '.make.state', '.nse_depinfo', '*~', '#*', '.#*', ',*',
+	 "_\$*", "*\$", '*.old', '*.bak', '*.BAK', '*.orig', '*.rej',
+	 '.del-*', '*.a', '*.olb', '*.o', '*.obj', '*.so', '*.exe',
+	 '*.Z', '*.elc', '*.ln', 'core');
+
+# 3-letter month names in POSIX locale, for fast date decoding
+my %months = (
+    "Jan" => 0,
+    "Feb" => 1,
+    "Mar" => 2,
+    "Apr" => 3,
+    "May" => 4,
+    "Jun" => 5,
+    "Jul" => 6,
+    "Aug" => 7,
+    "Sep" => 8,
+    "Oct" => 9,
+    "Nov" => 10,
+    "Dec" => 11
+);
 
 # print usage information and exit
 sub usage ()
 {
     print "Usage:\n" .
+	"  cvsu [OPTIONS] [FILE] ...\n" .
+	"Options:\n" .
 	"  --local		Disable recursion\n" .
 	"  --explain		Verbosely print status of files\n" .
 	"  --find		Emulate find - filenames only\n" .
 	"  --short		Don't print paths\n" .
 	"  --ignore		Don't read .cvsignore\n" .
 	"  --messages		List known file types and long messages\n" .
+	"  --nolinks		Disable recognizing hard and soft links\n" .
 	"  --types=[^]LIST	Print only file types [not] from LIST\n" .
 	"  --batch=COMMAND	Execute this command on files\n" .
 	"  --help		Print this usage information\n" .
@@ -70,19 +117,13 @@ sub list_messages ()
     exit 0;
 }
 
-# Initialize @standard_ignores
-# Also read $HOME/.cvsignore and append it to @standard_ignores
+# Initialize @common_ignores
+# Also read $HOME/.cvsignore and append it to @common_ignores
 sub init_ignores ()
 {
     my $HOME = $ENV{"HOME"};
-    # This list comes from the CVS manual.
-    @standard_ignores = ('RCS', 'SCCS', 'CVS', 'CVS.adm', 'RCSLOG',
-			 'cvslog.*', 'tags', 'TAGS', '.make.state',
-			 '.nse_depinfo', '*~', '#*', '.#*', ',*',
-			 "_\$*", "*\$", '*.old', '*.bak', '*.BAK',
-			 '*.orig', '*.rej', '.del-*', '*.a', '*.olb',
-			 '*.o', '*.obj', '*.so', '*.exe', '*.Z',
-			 '*.elc', '*.ln', 'core');
+
+    push @common_ignores, STANDARD_IGNORES;
 
     unless (defined($HOME)) {
 	return;
@@ -97,7 +138,7 @@ sub init_ignores ()
 	}
 
 	while (<CVSIGNORE>) {
-	    push (@standard_ignores, split);
+	    push (@common_ignores, split);
 	}
 
 	close (CVSIGNORE);
@@ -110,7 +151,7 @@ sub init_ignores ()
     }
 
     my @ignores_var = split (/ /, $CVSIGNOREENV);
-    push (@standard_ignores, @ignores_var);
+    push (@common_ignores, @ignores_var);
 
 }
 
@@ -169,90 +210,106 @@ sub file_status ($)
     }
 }
 
-# process one directory
-# Parameter 1: directory name
-sub process_dir ($);
-sub process_dir ($)
+# load entries from CVS/Entries and CVS/Entries.Log
+# Parameter 1: file name for CVS/Entries
+# Return: list of entries in the format used in CVS/Entries
+sub load_entries ($);
+sub load_entries ($)
 {
-    # 3-letter month names in POSIX locale
-    my %months = (
-	"Jan" => 0,
-	"Feb" => 1,
-	"Mar" => 2,
-	"Apr" => 3,
-	"May" => 4,
-	"Jun" => 5,
-	"Jul" => 6,
-	"Aug" => 7,
-	"Sep" => 8,
-	"Oct" => 9,
-	"Nov" => 10,
-	"Dec" => 11
-    );
+    my $entries_file = shift (@_);
+    my $entries_log_file = "$entries_file.Log";
+    my %ent = ();
 
-    # $file, $curr_dir, and $ignore_rx must be seen in file_status
-    $file = "";
-    $curr_dir = shift (@_);
-    $ignore_rx = '';
-
-    $curr_dir .= "/"
-	unless ( $curr_dir eq "" || $curr_dir =~ m{/$} );
-
-    my $real_curr_dir = $curr_dir eq "" ? "." : $curr_dir;
-
-    error ("$real_curr_dir is not a directory")
-	unless ( -d $real_curr_dir );
-
-    # Scan present files.
-    file_status (".");
-    my %found_files = ();
-    opendir (DIR, $real_curr_dir) ||
-	error ("couldn't open directory $real_curr_dir: $!");
-    foreach (readdir (DIR)) {
-	$found_files {$_} = 1;
+    unless (open (ENTRIES, "< $entries_file")) {
+	error ("couldn't open $entries_file: $!");
     }
-    closedir (DIR);
-
-    # Scan CVS/Entries.
-    my %entries = ();
-    my %subdirs = ();
-    my %removed = ();
-    open (ENTRIES, "< ${curr_dir}CVS/Entries")
-	|| error ("couldn't open ${curr_dir}CVS/Entries: $!");
     while (<ENTRIES>) {
-	if ( m{^D/([^/]+)/} ) {
-	    $subdirs{$1} = 1;
-	} elsif ( m{^/([^/]+)/([^/])[^/]*/([^/]*)/} ) {
-	    $entries{$1} = $3;
-	    $removed{$1} = 1
-		if $2 eq '-';
-	} else {
-	    error ("${curr_dir}CVS/Entries:$.: unrecognizable line")
-		unless m{D};
-	}
+	chomp;
+	$ent{$_} = 1;
     }
     close (ENTRIES);
 
-    # CVS/Entries.Log lists actions to be done in CVS/Entries
-    # Currently only adding and deleting directories is known to be safe
-    if ( open (ENTRIES, "< ${curr_dir}CVS/Entries.Log") ) {
+    if (open (ENTRIES, "< $entries_log_file")) {
 	while (<ENTRIES>) {
-	    if ( m{^A D/([^/]+)/} ) {
-		$subdirs{$1} = 1;
-	    } elsif ( m{^R D/([^/]+)/} ) {
-		delete $subdirs{$1};
+	    chomp;
+	    if ( m{^A (.+)} ) {
+		$ent{$1} = 1;
+	    } elsif ( m{^R (.+)} ) {
+		delete $ent{$1};
 	    } else {
 		# Note: "cvs commit" helps even when you are offline
-		error ("${curr_dir}CVS/Entries.Log:$.: unrecognizable line, " .
+		error ("$entries_log_file:$.: unrecognizable line, " .
 			"try \"cvs commit\"");
 	    }
 	}
 	close (ENTRIES);
     }
 
-    # It is intentional to list CVS before reading .cvsignore
-    $file = "CVS";
-    file_status ("C");
+    return keys %ent;
+}
+
+# process one directory
+# Parameter 1: directory name
+sub process_arg ($);
+sub process_arg ($)
+{
+    my $arg = shift (@_);
+    my %found_files = ();
+
+    # $file, $curr_dir, and $ignore_rx must be seen in file_status
+    local $file = "";
+    local $ignore_rx = "";
+    local $single_filename = 0;
+
+    if ( $arg eq "" or -d $arg ) {
+	$curr_dir = $arg;
+	my $real_curr_dir = $curr_dir eq "" ? "." : $curr_dir;
+
+	error ("$real_curr_dir is not a directory")
+	    unless ( -d $real_curr_dir );
+
+	# Scan present files.
+	file_status (".");
+	opendir (DIR, $real_curr_dir) ||
+	    error ("couldn't open directory $real_curr_dir: $!");
+	foreach (readdir (DIR)) {
+	    $found_files {$_} = 1;
+	}
+	closedir (DIR);
+    } else {
+	$single_filename = basename $arg;
+	$curr_dir = dirname $arg;
+	$found_files{$single_filename} = 1 if lstat $arg;
+    }
+
+    $curr_dir .= "/"
+	unless ( $curr_dir eq "" || $curr_dir =~ m{/$} );
+
+    # Scan CVS/Entries.
+    my %entries = ();
+    my %subdirs = ();
+    my %removed = ();
+
+    foreach ( load_entries ("${curr_dir}CVS/Entries") ) {
+	if ( m{^D/([^/]+)/} ) {
+	    $subdirs{$1} = SUBDIR_FOUND if !$single_filename;
+	} elsif ( m{^/([^/]+)/([^/])[^/]*/([^/]*)/} ) {
+	    if ( !$single_filename or $single_filename eq $1 ) {
+		$entries{$1} = $3;
+		$removed{$1} = 1
+		    if $2 eq '-';
+	    }
+	} elsif ( m{^D$} ) {
+	    next;
+	} else {
+	    error ("${curr_dir}CVS/Entries: unrecognizable line");
+	}
+    }
+
+    if ( $single_filename && !$entries{$single_filename} &&
+	 !$found_files{$single_filename} ) {
+	error ("nothing known about $arg");
+    }
 
     # Scan .cvsignore if any
     unless ($no_cvsignore) {
@@ -268,7 +325,7 @@ sub process_dir ($)
 	}
 
 	my ($iter);
-	foreach $iter (@ignore_list, @standard_ignores) {
+	foreach $iter (@ignore_list, @common_ignores) {
 	    if ($iter eq '!') {
 		$ignore_rx = ''
 	    } else {
@@ -296,25 +353,22 @@ sub process_dir ($)
     }
 
     foreach $file (sort keys %found_files) {
-	next if ($file eq 'CVS' || $file eq '.' || $file eq '..');
-	lstat ($curr_dir . $file); # Don't use stat() and -X on other files
-	my $is_link = 0;
-	eval {
-	    if (-l _) {
-		$is_link = 1;
-	    }
-	};
-	if ($is_link) {
+	next if ($file eq '.' || $file eq '..');
+	lstat ($curr_dir . $file) ||
+	    error ("lstat() failed on $curr_dir . $file");
+	if (! $nolinks && -l _) {
 	    file_status ("L");
 	} elsif (-d _) {
-	    if ($subdirs{$file}) {
-		$subdirs{$file} = 2;
+	    if ($file eq 'CVS') {
+		file_status ("C");
+	    } elsif ($subdirs{$file}) {
+		$subdirs{$file} = SUBDIR_CVS;
 	    } else {
 		file_status ("D"); # Unknown directory
 	    }
-	} elsif (! (-f _)) {
+	} elsif (! (-f _) && ! (-l _)) {
 	    file_status ("S"); # This must be something very special
-	} elsif ( (stat _) [3] > 1 ) {
+	} elsif (! $nolinks && (stat _) [3] > 1 ) {
 	    file_status ("H"); # Hard link
 	} elsif (! $entries{$file}) {
 	    file_status ("?");
@@ -342,11 +396,11 @@ sub process_dir ($)
     unless ($no_recurse) {
 	my $save_curr_dir = $curr_dir;
 	foreach $file (sort keys %subdirs) {
-	    if ($subdirs{$file} == 1) {
+	    if ($subdirs{$file} == SUBDIR_FOUND) {
 		$curr_dir = $save_curr_dir;
 		file_status ("X");
-	    } elsif ($subdirs{$file} == 2) {
-		process_dir ($save_curr_dir . $file)
+	    } elsif ($subdirs{$file} == SUBDIR_CVS) {
+		process_arg ($save_curr_dir . $file)
 	    }
 	}
     }
@@ -413,6 +467,7 @@ sub Main ()
     $find_mode = 0;		# Don't print status at all
     $short_print = 0;		# Print only filenames without path
     $no_cvsignore = 0;		# Ignore .cvsignore
+    $nolinks = 0;		# Do not test for soft- or hard-links
     my $want_msg = 0;		# List possible filetypes and exit
     my $want_help = 0;		# Print help and exit
     my $want_ver = 0;		# Print version and exit
@@ -426,6 +481,7 @@ sub Main ()
 	"short"	   => \$short_print,
 	"ignore"   => \$no_cvsignore,
 	"messages" => \$want_msg,
+	"nolinks"  => \$nolinks,
 	"help"     => \$want_help,
 	"version"  => \$want_ver
     );
@@ -447,7 +503,7 @@ sub Main ()
     }
 
     foreach (@ARGV) {
-	process_dir ($_);
+	process_arg ($_);
     }
 
     if ($#batch_list >= 0) {
